package drones.dispatcher;

import drones.I18N;
import drones.drone.Drone;
import drones.geo.Point;
import drones.geo.TubeStation;
import drones.messages.EventBus;
import drones.messages.EventBus.Listener;
import drones.messages.Message;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.BufferedReader;
import java.io.IOException;
import java.text.DecimalFormat;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.*;

/**
 * A Drone dispatcher for controlling {@link Drone}'s
 *
 * @author John Ahlroos
 */
public class Dispatcher implements Runnable, Listener<Message> {

    public static final DateTimeFormatter DATE_FORMAT = DateTimeFormatter.ofPattern(I18N.get("date.format"));

    private static final DecimalFormat REPORT_NUMBER_FORMAT = new DecimalFormat(I18N.get("number.format"));
    private static final Logger LOGGER = LoggerFactory.getLogger("Dispatcher");
    private static final Logger REPORT_LOGGER = LoggerFactory.getLogger("Report");

    private final LocalDateTime shutdownTime;
    private final Map<Long, BufferedReader> dataReaders = new HashMap<>();
    private final double simulatedTimeFactor;
    private LocalDateTime currentTime;

    private final EventBus eventBus;
    private final List<TubeStation> tubeStations = new ArrayList<>();

    private final Set<String> tubeStationsFound = new HashSet<>();
    private int tuneStationReports = 0;
    private final Map<Long, Long> waypointsSent = new HashMap<>();

    /**
     * Creates a new dispatcher
     *
     * @param currentTime         the current time of the dispatcher
     * @param shutdownTime        the time when the dispatcher should terminate
     * @param simulatedTimeFactor used for controlling simulated time. Should be between 0-1.
     */
    public Dispatcher(LocalDateTime currentTime, LocalDateTime shutdownTime, double simulatedTimeFactor) {
        this.currentTime = currentTime;
        this.shutdownTime = shutdownTime;
        this.simulatedTimeFactor = Math.max(0.0001, simulatedTimeFactor);
        this.eventBus = new EventBus(simulatedTimeFactor);
        this.eventBus.register(this);
        LOGGER.info(I18N.get("dispatcher.running.stats.1"));
        LOGGER.info(I18N.get("dispatcher.running.stats.2"), this.currentTime);
        LOGGER.info(I18N.get("dispatcher.running.stats.3"), this.shutdownTime);
        LOGGER.info(I18N.get("dispatcher.running.stats.4"), this.simulatedTimeFactor);
        LOGGER.info("");
    }

    /**
     * Register tube stations to report on.
     * <p>
     * The format is:
     * "<Station name>",<latitude>,<longitude>
     *
     * @param reader A reader referencing a file and another data stream
     * @throws IOException if reading the file fails
     */
    public void registerTubeStations(BufferedReader reader) throws IOException {
        try (reader) {
            for (String line = reader.readLine(); line != null; line = reader.readLine()) {
                var scanner = new Scanner(line).useDelimiter(",");
                var name = scanner.next().replace("\"", "");
                var lat = Double.parseDouble(scanner.next().replace("\"", ""));
                var lon = Double.parseDouble(scanner.next().replace("\"", ""));
                var tb = new TubeStation(new Point(lat, lon, null), name);
                LOGGER.info("Registered Tube Station: {}", tb);
                tubeStations.add(tb);
            }
        }
    }

    /**
     * Registers a new drone with the dispatcher.
     *
     * @param id     the drone id
     * @param reader the data source reader, can point to a file or another data stream.
     *               The format is:
     *               <drone id>,"<latitude>","<longitude>","<time>"
     *               The time format is:
     *               2011-03-22 07:49:10
     * @return returns a representation of the drone
     */
    public Runnable registerDroneWithReader(long id, BufferedReader reader) {
        dataReaders.put(id, reader);
        var drone = new Drone(id, currentTime, simulatedTimeFactor, eventBus, tubeStations);
        eventBus.register(drone);
        LOGGER.info(I18N.get("dispatcher.drone.registration"), drone.getId());
        return drone;
    }

    @Override
    public void run() {
        LOGGER.info(I18N.get("dispatcher.started"));

        while (dataReaders.size() > 0) {
            if (currentTime.isEqual(shutdownTime)) {
                LOGGER.info(I18N.get("dispatcher.send.shutdown.signals"), dataReaders.size());
                for (Long droneId : dataReaders.keySet()) {
                    LOGGER.info(I18N.get("dispatcher.event.drone.shutdown"), droneId);
                    eventBus.push(new Message.ShutdownSignal(droneId, droneId, shutdownTime));
                }
            }

            if (Thread.currentThread().isInterrupted()) {
                LOGGER.info(I18N.get("dispatcher.shutdown"));
                break;
            }

            incrementTime();
        }

        LOGGER.info(I18N.get("dispatcher.terminating"));
        LOGGER.info(I18N.get("dispatcher.termination.stats.1"), tubeStationsFound);
        LOGGER.info(I18N.get("dispatcher.termination.stats.2"), tuneStationReports);
        LOGGER.info(I18N.get("dispatcher.termination.stats.3"));
        waypointsSent.forEach((k, v) -> LOGGER.info(I18N.get("dispatcher.termination.stats.4"), k, v));
    }

    @Override
    public void onEvent(Message message) {

        // No more drones
        if (dataReaders.isEmpty()) {
            return;
        }

        // Ensure we are only processing our messages
        if (message.targetId() != Message.DISPATCHER_ID) {
            return;
        }

        // Received shutdown confirmation from Drone, cleanup resources
        if (message instanceof Message.ShutdownSignal shutdown) {
            try {
                if (dataReaders.containsKey(shutdown.droneId())) {
                    dataReaders.get(shutdown.droneId()).close();
                    dataReaders.remove(shutdown.droneId());
                }
            } catch (IOException e) {
                LOGGER.warn(I18N.get("dispatcher.event.drone.data.stream.error"), shutdown.droneId());
            }
        }

        // Is drone request route waypoints?
        if (message instanceof Message.DroneCapacity droneCapacity) {
            LOGGER.info(I18N.get("dispatcher.event.drone.capacity.request"), droneCapacity.droneId(), droneCapacity.capacity());
            updateDroneRoute(droneCapacity.droneId(), droneCapacity.capacity());
            return;
        }

        // Has drone sent a traffic condition?
        if (message instanceof Message.TrafficCondition condition) {
            tubeStationsFound.add(condition.tubeStation());
            tuneStationReports++;
            REPORT_LOGGER.info(I18N.get("dispatcher.event.drone.traffic.report"),
                    condition.tubeStation(),
                    condition.time().toLocalTime(),
                    condition.condition().toString(),
                    condition.droneId(), REPORT_NUMBER_FORMAT.format(condition.speed() * 3.6),
                    REPORT_NUMBER_FORMAT.format(condition.distanceToStation()));
            return;
        }
    }

    /**
     * Reads the next waypoint from the data source.
     *
     * @param id The id of the drone
     * @return If no waypoint can be read, either by reaching the end-of-file or by an error, an empty Optional is
     * returned. Otherwise the read point is returned.
     */
    private Optional<Point> readNextPosition(long id) {
        try {
            var reader = dataReaders.get(id);
            if (reader == null) {
                return Optional.empty();
            }

            var line = reader.readLine();
            if (line == null) {
                return Optional.empty();
            }

            var scanner = new Scanner(line).useDelimiter(",");
            if (scanner.nextLong(10) != id) {
                throw new IllegalStateException(I18N.get("dispatcher.data.file.id.mismatch"));
            }

            var lat = Double.parseDouble(scanner.next().replace("\"", ""));
            var lon = Double.parseDouble(scanner.next().replace("\"", ""));
            var time = LocalDateTime.parse(scanner.next().replace("\"", ""), DATE_FORMAT);

            if (time.isAfter(shutdownTime)) {
                // Ignore points after shutdown time
                return Optional.empty();
            }

            return Optional.of(new Point(lat, lon, time));
        } catch (IOException e) {
            LOGGER.warn(I18N.get("dispatcher.data.file.stream.error"));
            return Optional.empty();
        }
    }

    /**
     * Updates a drone with new waypoints if there is capacity
     *
     * @param droneId      the id of the drone to update
     * @param fillCapacity the amount of waypoints to send
     */
    private void updateDroneRoute(long droneId, int fillCapacity) {
        var addedWaypoints = 0;
        while (fillCapacity > 0) {
            var wp = readNextPosition(droneId);
            if (wp.isEmpty()) {
                break;
            }
            eventBus.push(new Message.PointData(droneId, wp.get()));
            waypointsSent.computeIfPresent(droneId, (k, v) -> ++v);
            waypointsSent.putIfAbsent(droneId, 1L);
            addedWaypoints++;
            fillCapacity--;
        }
        if (addedWaypoints > 0) {
            LOGGER.info(I18N.get("dispatcher.waypoints.sent"), addedWaypoints, droneId);
        }
    }

    /**
     * Simulates latency of the dispatcher
     * <p>
     * Controlled by the {@link #simulatedTimeFactor} property. If using real-time (factor 1) then there is a 10
     * second latency.
     */
    private void incrementTime() {
        try {
            currentTime = currentTime.plusSeconds(1);
            Thread.sleep(Math.round(simulatedTimeFactor * 1000));
        } catch (InterruptedException e) {
            LOGGER.debug(I18N.get("dispatcher.thread.sleep.error"));
        }
    }
}
