package drones.drone;

import drones.I18N;
import drones.geo.Point;
import drones.geo.TubeStation;
import drones.messages.EventBus;
import drones.messages.EventBus.Listener;
import drones.messages.Message;
import drones.messages.Message.TrafficCondition.Condition;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.time.LocalDateTime;
import java.util.*;
import java.util.concurrent.ConcurrentLinkedDeque;

/**
 * Drone implementation to be controlled by a {@link drones.dispatcher.Dispatcher}
 *
 * @author John Ahlroos
 */
public class Drone implements Runnable, Listener<Message> {

    private static final int ROUTE_CAPACITY = 10;
    private static final int REQUEST_WAYPOINT_THRESHOLD = 3; // 30% capacity remaining
    public static final int TUBE_STATION_RANGE = 350;

    private final long id;
    private double speed = 0;
    private Point position;
    private final Deque<Point> route = new ConcurrentLinkedDeque<>();
    private DroneState state = DroneState.HOLDING;
    private LocalDateTime currentTime;
    private final double simulatedTimeFactor;
    private final EventBus eventBus;
    private final Collection<TubeStation> tubeStations;
    private final Set<TubeStation> tubeStationsInRange = new HashSet<>();
    private final Set<TubeStation> tubeStationsFound = new HashSet<>();
    private LocalDateTime shutdownTime;

    private int waypointsVisited = 0;
    private double longestDistanceTraveled = 0;
    private double fastestSpeed = 0;

    /**
     * Creates a new drone
     *
     * @param id                  the id of the drone. Must be unique
     * @param currentTime         the current time
     * @param simulatedTimeFactor time factor to simulate time. Should be between 0-1.
     * @param eventBus            the event bus for sending and recieving messages
     * @param tubeStations        the tube stations to report on
     */
    public Drone(long id, LocalDateTime currentTime, double simulatedTimeFactor, EventBus eventBus,
                 Collection<TubeStation> tubeStations) {
        this.id = id;
        this.currentTime = currentTime;
        this.simulatedTimeFactor = Math.max(0.0001, simulatedTimeFactor);
        this.eventBus = eventBus;
        this.tubeStations = tubeStations;
        this.position = new Point(0, 0, currentTime);
        getLogger().info(I18N.get("drone.running.stats.1"));
        getLogger().info(I18N.get("drone.running.stats.2"), this.currentTime);
        getLogger().info(I18N.get("drone.running.stats.3"), this.simulatedTimeFactor);
        getLogger().info("");
    }

    /**
     * Get the id of the drone. Must be unique.
     *
     * @return the drone id
     */
    public long getId() {
        return id;
    }

    /**
     * Get the current position of the drone.
     * <p>
     * If the drone has not yet recieved any waypoints then this is <code>null</code>. This will be set initially to
     * the first waypoint and then updated as the drone flies.
     * </p>
     *
     * @return the current position
     */
    public Point getPosition() {
        return position;
    }

    /**
     * Get how many waypoints remain on the current route. Might be updated by the dispatcher.
     *
     * @return the amount of waypoints remaining
     */
    public int getWaypointsRemaining() {
        return route.size();
    }

    /**
     * The current state of the drone.
     *
     * <ul>
     *     <li>{@link DroneState#HOLDING}: The drone is currently holding at the same position</li>
     *     <li>{@link DroneState#MOVING}: The drone is currently moving to a waypoint</li>
     *     <li>{@link DroneState#TERMINATED}: The drone has terminated operation. No more waypoints will be
     *     processed</li>
     * </ul>
     *
     * @return the current drone state.
     */
    public DroneState getState() {
        return state;
    }

    /**
     * Returns how many waypoints have been visited by the drone
     *
     * @return the number of waypoints visited
     */
    public int getWaypointsVisited() {
        return waypointsVisited;
    }

    /**
     * Returns which tube stations have been found while operating
     *
     * @return the names of the tube stations found
     */
    public Set<TubeStation> getTubeStationsFound() {
        return Collections.unmodifiableSet(tubeStationsFound);
    }

    @Override
    public void run() {
        getLogger().info(I18N.get("drone.started"));

        // Request initial waypoints
        requestWaypoints();

        while (state != DroneState.TERMINATED) {

            // Received termination signal
            if (Thread.currentThread().isInterrupted()) {
                getLogger().info(this.toString());
                break;
            }

            // Shutdown
            if (shutdownTime != null && (currentTime.isEqual(shutdownTime) || currentTime.isAfter(shutdownTime))) {
                if (route.isEmpty()) {
                    setState(DroneState.TERMINATED);
                    getLogger().info(this.toString());
                    break;
                }
            }

            // Log current location and drone details
            getLogger().info(this.toString());

            // Check if we have any more waypoints, if not hold position
            var nextWaypoint = route.peekFirst();
            if (nextWaypoint == null) {
                getLogger().info(I18N.get("drone.holding.position"));
                setState(DroneState.HOLDING);
                speed = 0.0;
                incrementTime();
                continue;
            }

            // Check if we received our first waypoint, adjust dummy position
            if (position.latitude() == 0 && nextWaypoint.time().isBefore(currentTime)) {
                getLogger().info(I18N.get("drone.holding.position"));
                setState(DroneState.HOLDING);
                speed = 0.0;
                incrementTime();
                continue;
            } else if (position.latitude() == 0) {
                position = Point.adjustTime(nextWaypoint, position.time());
            }

            // Check for tube stations
            tubeStationsInRange.removeIf(tb -> !tb.point().isInRange(position, TUBE_STATION_RANGE));
            tubeStations.stream()
                    .filter(tb -> !tubeStationsInRange.contains(tb))
                    .filter(tb -> tb.point().isInRange(position, TUBE_STATION_RANGE))
                    .peek(tubeStationsInRange::add)
                    .peek(tubeStationsFound::add)
                    .forEach(tb -> eventBus.push(new Message.TrafficCondition(
                            Message.DISPATCHER_ID, getId(), tb.name(),
                            position.time(), speed,
                            Condition.values()[new Random().nextInt(Condition.values().length)],
                            position.distanceTo(tb.point()))));

            // Check if we are at the destination waypoint
            if (nextWaypoint.equals(position)) {
                setState(DroneState.HOLDING);
                speed = 0.0;
                currentTime = nextWaypoint.time();
                waypointsVisited++;
                getLogger().info(I18N.get("drone.arrived.at.waypoint"), nextWaypoint);
                position = route.removeFirst();
                requestWaypoints();
                incrementTime();
                continue;
            }

            // Calculate new speed if we are departing to a new waypoint
            var distanceToNextWaypoint = position.distanceTo(nextWaypoint);
            longestDistanceTraveled = Math.max(longestDistanceTraveled, distanceToNextWaypoint);
            if (state == DroneState.HOLDING) {
                var timeToNextWaypoint = position.timeTo(nextWaypoint);
                speed = distanceToNextWaypoint / timeToNextWaypoint.getSeconds();
                getLogger().debug(I18N.get("drone.speed.calculation"), timeToNextWaypoint.getSeconds(), distanceToNextWaypoint);
            }

            // Prevent overshooting the waypoint by reducing speed as we approach the waypoint
            if (speed > distanceToNextWaypoint) {
                speed = distanceToNextWaypoint;
                getLogger().info(I18N.get("drone.speed.reduced"), speed);
            }

            // Start moving to the new waypoint
            fastestSpeed = Math.max(fastestSpeed, speed);
            position = position.moveTowards(nextWaypoint, speed, 1);
            setState(DroneState.MOVING);
            getLogger().debug(I18N.get("drone.next.waypoint.leg"), position);
            incrementTime();
        }

        getLogger().info(I18N.get("drone.shutdown"), position);
        eventBus.push(new Message.ShutdownSignal(Message.DISPATCHER_ID, id, shutdownTime));

        getLogger().info(I18N.get("drone.termination.stats.1"), waypointsVisited);
        getLogger().info(I18N.get("drone.termination.stats.2"), tubeStationsFound);
        getLogger().info(String.format(I18N.get("drone.termination.stats.3"), longestDistanceTraveled / 1000.0));
        getLogger().info(String.format(I18N.get("drone.termination.stats.4"), fastestSpeed * 3.6));
    }

    private void requestWaypoints() {
        var capacity = availableRouteCapacity();
        if (capacity > ROUTE_CAPACITY - REQUEST_WAYPOINT_THRESHOLD) {
            eventBus.push(new Message.DroneCapacity(Message.DISPATCHER_ID, getId(), capacity));
            getLogger().info(I18N.get("drone.new.route.request"), capacity);
        }
    }

    private void setState(DroneState state) {
        if (this.state == DroneState.TERMINATED) {
            return;
        }
        this.state = state;
    }

    @Override
    public String toString() {
        var lat = position.latitude();
        var lon = position.longitude();
        var time = position.time();
        return String.format(I18N.get("drone.position"), speed, state, lat, lon, route.size(), waypointsVisited,
                time.toLocalTime());
    }

    @Override
    public void onEvent(Message message) {

        // Check that the message is for this drone
        if (message.targetId() != id) {
            return;
        }

        // Check for signals
        if (message instanceof Message.ShutdownSignal signal) {
            getLogger().info(I18N.get("drone.event.signal"), signal);
            shutdownTime = signal.time();
            getLogger().info(I18N.get("drone.event.signal.shutdown"));
            return;
        }

        // Check for new waypoints
        if (message instanceof Message.PointData data) {
            var point = data.point();
            if (availableRouteCapacity() > 0) {
                getLogger().info(I18N.get("drone.event.new.waypoint"), point);
                route.add(point);
            } else {
                getLogger().warn(I18N.get("drone.event.new.waypoint.cache.full"), point);
            }
            return;
        }
    }

    /**
     * Get the amount of points we can load in the waypoint cache
     *
     * @return number of points we can load
     */
    private int availableRouteCapacity() {
        return ROUTE_CAPACITY - route.size();
    }

    /**
     * Get the named logger for the drone (allows us to split the logs into different files)
     *
     * @return the logger
     */
    private Logger getLogger() {
        return LoggerFactory.getLogger("Drone." + id);
    }

    /**
     * Simulates latency and delay while flying. Allows us to model in real-time the fight path
     */
    private void incrementTime() {
        try {
            currentTime = currentTime.plusSeconds(1);
            position = Point.adjustTime(position, currentTime);
            Thread.sleep(Math.round(simulatedTimeFactor * 1000));
        } catch (InterruptedException e) {
            getLogger().debug(I18N.get("drone.thread.sleep.error"));
        }
    }
}
