package drones.geo;

import drones.I18N;

import java.time.Duration;
import java.time.LocalDateTime;
import java.util.Objects;

/**
 * Represents a point in the world.
 *
 * @author John Ahlroos
 */
public record Point(double latitude, double longitude, LocalDateTime time) {

    private static final double EARTH_RADIUS = 6371.01;
    private static final double ACCURACY = 0.00001;

    /**
     * Measures the distance between this point and another point.
     *
     * @param point the point to measure distance to
     * @return the distance in meters
     */
    public double distanceTo(Point point) {
        var lat1 = Math.toRadians(latitude());
        var lon1 = Math.toRadians(longitude());
        var lat2 = Math.toRadians(point.latitude());
        var lon2 = Math.toRadians(point.longitude());
        double earthRadius = EARTH_RADIUS * 1000; //Meters
        return earthRadius * Math.acos(Math.sin(lat1) * Math.sin(lat2) + Math.cos(lat1) * Math.cos(lat2) * Math.cos(lon1 - lon2));
    }

    /**
     * Checks if another point is withing a certain bounds
     *
     * @param point  the point to check
     * @param meters the range to check for
     * @return <code>true</code> if it is within range, <code>false</code> otherwise.
     */
    public boolean isInRange(Point point, double meters) {
        return distanceTo(point) <= meters;
    }

    /**
     * The amount of time between two point timestamps
     *
     * @param point the other point to compare to
     * @return the duraction between the points
     */
    public Duration timeTo(Point point) {
        var time = Duration.between(time(), point.time().plusSeconds(1)).abs();
        return time.isZero() ? Duration.ofSeconds(1) : time;
    }

    /**
     * The bearing, measured in radians, to another points
     *
     * @param point the other point to point to
     * @return the bearing in radians
     */
    public double bearingTo(Point point) {
        var dL = point.longitude() - longitude();
        var x = Math.cos(point.latitude()) * Math.sin(dL);
        var y = Math.cos(latitude()) * Math.sin(point.latitude()) -
                Math.sin(latitude()) * Math.cos(point.latitude()) * Math.cos(dL);
        return Math.atan2(x, y);
    }

    /**
     * Calculates what the next point would be if we would move (meters) amount
     * for (seconds) toward another point.
     *
     * @param destination The destination point to head for
     * @param meters      the amount of meters to travel
     * @param seconds     the amount of time it should take
     * @return the next point we would reach
     */
    public Point moveTowards(Point destination, double meters, long seconds) {
        if (meters <= 0) {
            throw new IllegalArgumentException(I18N.get("point.meters.positive.value"));
        }
        if (seconds <= 0) {
            throw new IllegalArgumentException(I18N.get("point.seconds.positive.value"));
        }
        var brng = bearingTo(destination);
        var d = meters / 1000.0;
        var lat1 = Math.toRadians(latitude());
        var lon1 = Math.toRadians(longitude());
        var lat2 = Math.asin(Math.sin(lat1) * Math.cos(d / EARTH_RADIUS) +
                Math.cos(lat1) * Math.sin(d / EARTH_RADIUS) * Math.cos(brng));
        var lon2 = lon1 + Math.atan2(Math.sin(brng) * Math.sin(d / EARTH_RADIUS) * Math.cos(lat1),
                Math.cos(d / EARTH_RADIUS) - Math.sin(lat1) * Math.sin(lat2));
        return new Point(Math.toDegrees(lat2), Math.toDegrees(lon2), time());
    }

    /**
     * Returns a new point where the time has been adjusted to the new time
     *
     * @param point the point to use as basis
     * @param time  the time to set
     * @return a new point using the old point but with changed time
     */
    public static Point adjustTime(Point point, LocalDateTime time) {
        return new Point(point.latitude(), point.longitude(), time);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        var point = (Point) o;
        var latDiff = Math.abs(point.latitude - latitude);
        var lonDiff = Math.abs(point.longitude - longitude);
        return latDiff < ACCURACY && lonDiff < ACCURACY;
    }

    @Override
    public int hashCode() {
        return Objects.hash(latitude, longitude);
    }

    @Override
    public String toString() {
        return String.format("(lat=%f, lon=%f, time=%s)", latitude, longitude, time);
    }
}
