package drones.messages;

import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;

/**
 * Represent a simple event bus between the {@link drones.drone.Drone} and the {@link drones.dispatcher.Dispatcher}
 *
 * @author John Ahlroos
 */
public class EventBus {

    private final List<Listener<Message>> listeners = new CopyOnWriteArrayList<>();
    private final double simulatedTimeFactor;

    public EventBus(double simulatedTimeFactor) {
        this.simulatedTimeFactor = simulatedTimeFactor;
    }

    /**
     * Push a new message to the event bus
     *
     * @param message the message to send
     */
    public void push(Message message) {
        latency();
        listeners.forEach(l -> l.onEvent(message));
    }

    /**
     * Register a receiver for the messages
     *
     * @param listener the listener
     */
    public void register(Listener<Message> listener) {
        listeners.add(listener);
    }

    /**
     * Simulates a slight latency in the event bus (messages moving over the air)
     */
    private void latency() {
        try {
            Thread.sleep(Math.round(simulatedTimeFactor * 300));
        } catch (InterruptedException e) {
            // Ignore
        }
    }

    /**
     * Listener interface for receiving messages from the event bus
     *
     * @param <Message> messsage on the buss
     */
    @FunctionalInterface
    public interface Listener<Message> {

        /**
         * Triggered when a new message is published on the bus
         *
         * @param message the message that was published
         */
        void onEvent(Message message);
    }
}
