package drones.dispatcher;

import drones.drone.Drone;
import drones.drone.DroneState;
import org.junit.jupiter.api.Test;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.time.LocalDateTime;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;

public class DispatcherTest {

    @Test
    public void setupDrone() {
        var start = LocalDateTime.parse("2011-03-22 07:55:20", Dispatcher.DATE_FORMAT);
        var end = LocalDateTime.parse("2011-03-22 07:55:40", Dispatcher.DATE_FORMAT);
        var route = """
                5937,"51.476105","-0.100224","2011-03-22 07:55:26"
                5937,"51.475967","-0.100368","2011-03-22 07:55:40"
                """;

        var dispatcher = new Dispatcher(start, end, 0);
        var drone = (Drone) dispatcher.registerDroneWithReader(5937, new BufferedReader(new StringReader(route)));
        assertEquals(drone.getPosition().time(), start);
        assertEquals(0, drone.getWaypointsRemaining());
        assertEquals(5937, drone.getId());
    }

    @Test
    public void pushDataToDrone() throws InterruptedException {
        var start = LocalDateTime.parse("2011-03-22 07:55:20", Dispatcher.DATE_FORMAT);
        var end = LocalDateTime.parse("2011-03-22 07:56:00", Dispatcher.DATE_FORMAT);
        var route = """
                5937,"51.476105","-0.100224","2011-03-22 07:55:26"
                5937,"51.475967","-0.100368","2011-03-22 07:55:27"
                5937,"51.476021","-0.100246","2011-03-22 07:55:28"
                5937,"51.476051","-0.100078","2011-03-22 07:55:29"
                5937,"51.476009","-0.099922","2011-03-22 07:55:30"
                5937,"51.476044","-0.099775","2011-03-22 07:55:31"
                5937,"51.476074","-0.099968","2011-03-22 07:55:42"
                5937,"51.476086","-0.100047","2011-03-22 07:55:43"
                5937,"51.476074","-0.100123","2011-03-22 07:55:44"
                5937,"51.476089","-0.10019","2011-03-22 07:55:45"
                5937,"51.476112","-0.100246","2011-03-22 07:55:46"
                5937,"51.476112","-0.100264","2011-03-22 07:55:47"
                5937,"51.476116","-0.10028","2011-03-22 07:55:48"
                5937,"51.476112","-0.100356","2011-03-22 07:55:49"
                5937,"51.476135","-0.100378","2011-03-22 07:55:50"
                5937,"51.476154","-0.100392","2011-03-22 07:55:51"
                5937,"51.476189","-0.100396","2011-03-22 07:55:52"
                5937,"51.476219","-0.100387","2011-03-22 07:55:53"
                5937,"51.476257","-0.100379","2011-03-22 07:55:54"
                """;

        var dispatcher = new Dispatcher(start, end, 0);

        var drone = (Drone) dispatcher.registerDroneWithReader(5937, new BufferedReader(new StringReader(route)));
        var droneRunner = new Thread(drone);

        droneRunner.start();
        dispatcher.run();
        droneRunner.join();

        assertEquals(route.lines().count(), drone.getWaypointsVisited());
        assertEquals(0, drone.getWaypointsRemaining());
        assertEquals(DroneState.TERMINATED, drone.getState());
    }

    @Test
    public void registerTubeStations() throws IOException, InterruptedException {
        var start = LocalDateTime.parse("2011-03-22 07:55:20", Dispatcher.DATE_FORMAT);
        var end = LocalDateTime.parse("2011-03-22 07:56:00", Dispatcher.DATE_FORMAT);
        var dispatcher = new Dispatcher(start, end, 0);
        var stations = """
                "Acton Town",51.503071,-0.280303
                "Aldgate",51.514342,-0.075627
                "Aldgate East",51.51503,-0.073162
                "All Saints (DLR)",51.510477,-0.012625
                """;
        dispatcher.registerTubeStations(new BufferedReader(new StringReader(stations)));

        var route = """
                5937,"51.476105","-0.100224","2011-03-22 07:55:26"
                5937,"51.51503","-0.073162","2011-03-22 07:55:27"
                5937,"51.476021","-0.100246","2011-03-22 07:55:28"
                """;
        var drone = (Drone) dispatcher.registerDroneWithReader(5937, new BufferedReader(new StringReader(route)));
        var droneRunner = new Thread(drone);

        droneRunner.start();
        dispatcher.run();
        droneRunner.join();

        var foundTubeStations = drone.getTubeStationsFound();
        assertTrue(foundTubeStations.stream().anyMatch(tb -> tb.name().equals("Aldgate")));
        assertTrue(foundTubeStations.stream().anyMatch(tb -> tb.name().equals("Aldgate East")));
    }
}
