package drones.drone;

import drones.dispatcher.Dispatcher;
import drones.geo.Point;
import drones.messages.EventBus;
import drones.messages.Message;
import org.junit.jupiter.api.Test;

import java.time.LocalDateTime;
import java.util.Collections;
import java.util.List;

import static org.junit.jupiter.api.Assertions.assertEquals;

public class DroneTest {

    @Test
    public void setupDrone() {
        var points = generatePoints();
        var eventBus = new EventBus(0);

        var drone = new Drone(1L, getStartTime(), 1, eventBus, Collections.emptyList());
        eventBus.register(drone);

        eventBus.push(new Message.PointData(drone.getId(), points.get(0)));
        eventBus.push(new Message.PointData(drone.getId(), points.get(1)));
        eventBus.push(new Message.PointData(drone.getId(), points.get(2)));

        assertEquals(3, drone.getWaypointsRemaining());
        assertEquals(0, drone.getWaypointsVisited());
    }

    @Test
    public void initialWaypointIsSetAsPosition() throws InterruptedException {
        var points = generatePoints();
        var eventBus = new EventBus(0);

        var drone = new Drone(1L, getStartTime(), 0, eventBus, Collections.emptyList());
        eventBus.register(drone);

        var droneThread = new Thread(drone);
        droneThread.start();

        eventBus.push(new Message.PointData(drone.getId(), points.get(0)));
        eventBus.push(new Message.ShutdownSignal(drone.getId(), drone.getId(), getEndTime()));

        droneThread.join();

        assertEquals(drone.getPosition(), points.get(0));
        assertEquals(drone.getWaypointsVisited(), 1);
        assertEquals(drone.getWaypointsRemaining(), 0);
    }

    @Test
    public void travelBetweenWaypoints() throws InterruptedException {
        var points = generatePoints();
        var eventBus = new EventBus(0);

        var drone = new Drone(1L, getStartTime(), 0, eventBus, Collections.emptyList());
        eventBus.register(drone);

        var droneThread = new Thread(drone);
        droneThread.start();

        eventBus.push(new Message.PointData(drone.getId(), points.get(0)));
        eventBus.push(new Message.PointData(drone.getId(), points.get(1)));
        eventBus.push(new Message.PointData(drone.getId(), points.get(2)));
        eventBus.push(new Message.ShutdownSignal(drone.getId(), drone.getId(), getEndTime()));

        droneThread.join();

        assertEquals(DroneState.TERMINATED, drone.getState());
        assertEquals(getEndTime(), drone.getPosition().time());
        assertEquals(0, drone.getWaypointsRemaining());
        assertEquals(3, drone.getWaypointsVisited());
    }

    private List<Point> generatePoints() {
        var point1 = new Point(51.476105, -0.100224, LocalDateTime.parse("2011-03-22 07:55:26", Dispatcher.DATE_FORMAT));
        var point2 = new Point(51.475967, -0.100368, LocalDateTime.parse("2011-03-22 07:55:30", Dispatcher.DATE_FORMAT));
        var point3 = new Point(51.476021, -0.100246, LocalDateTime.parse("2011-03-22 07:55:34", Dispatcher.DATE_FORMAT));
        return List.of(point1, point2, point3);
    }

    private LocalDateTime getStartTime() {
        return LocalDateTime.parse("2011-03-22 07:55:20", Dispatcher.DATE_FORMAT);
    }

    private LocalDateTime getEndTime() {
        return LocalDateTime.parse("2011-03-22 07:55:40", Dispatcher.DATE_FORMAT);
    }
}
