package simulation;

import drones.dispatcher.Dispatcher;
import picocli.CommandLine;
import picocli.CommandLine.Command;
import picocli.CommandLine.Help.Visibility;
import picocli.CommandLine.ITypeConverter;
import picocli.CommandLine.Option;
import picocli.CommandLine.Parameters;

import java.nio.file.Files;
import java.nio.file.Path;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

/**
 * Simulates flying drones by using data form the ./data directory
 * <p>
 * Creates log files to the logs/ directory:
 * - dispatcher.log: Logs from the dispatcher controlling the drones
 * - drone-<id>.log: Logs from the drones
 * <p>
 * The output of the simulation that is printed in the console is a report for traffic conditions of the provided
 * tube stations. Every line represents the condition of the traffic at a certain point in time and from a certain
 * waypoint viewpoint.
 */

@Command(name = "drone-simulator", mixinStandardHelpOptions = true,
        version = "1.0", resourceBundle = Simulation.RESOURCE_BUNDLE, showDefaultValues = true)
public class Simulation implements Callable<Integer> {

    public static final String RESOURCE_BUNDLE = "simulation-messages";

    @Parameters(arity = "1", paramLabel = "DRONES", showDefaultValue = Visibility.NEVER)
    private List<Long> droneIds;

    @Option(names = {"-d", "--data-dir"}, defaultValue = ".")
    private Path dataDir;

    @Option(names = {"-t", "--tube-stations"}, defaultValue = "./tube.csv")
    private Path tubeStationsFile;

    @Option(names = {"-b", "--start-time"}, defaultValue = "2011-03-22 07:47:00", converter = TimeConverter.class)
    private LocalDateTime currentTime;

    @Option(names = {"-e", "--shut-down-time"}, defaultValue = "2011-03-22 08:10:00", converter = TimeConverter.class)
    private LocalDateTime shutDownTime;

    @Option(names = {"-p", "--simulation-speed"}, defaultValue = "0.0")
    private double simulationSpeed;

    @Override
    public Integer call() throws Exception {
        System.out.println(I18N.get("starting.message"));
        System.out.println();

        // Use default location for tube file (in the data directory) if not explicitly set
        if (!tubeStationsFile.toFile().exists()) {
            tubeStationsFile = dataDir.resolve(Path.of("tube.csv"));
        }

        final var executor = Executors.newFixedThreadPool(droneIds.size() + 1);
        try {

            // Setup dispatcher
            var dispatcher = new Dispatcher(currentTime, shutDownTime, simulationSpeed);
            dispatcher.registerTubeStations(Files.newBufferedReader(tubeStationsFile));

            // Setup Drones
            var drones = new ArrayList<Runnable>();
            for (long id : droneIds) {
                var droneData = Files.newBufferedReader(dataDir.resolve(Path.of(id + ".csv")));
                drones.add(dispatcher.registerDroneWithReader(id, droneData));
            }

            // Power them on
            executor.execute(dispatcher);
            drones.forEach(executor::execute);

        } finally {
            executor.shutdown();
            executor.awaitTermination(20, TimeUnit.MINUTES);
        }
        System.out.println(I18N.get("ended.message"));
        return 0;
    }

    /**
     * Main entrypoint of the simulation
     *
     * @param args program arguments
     */
    public static void main(String... args) {
        int exitCode = new CommandLine(new Simulation()).execute(args);
        System.exit(exitCode);
    }

    /**
     * Set the date formats for input times
     */
    static class TimeConverter implements ITypeConverter<LocalDateTime> {
        public LocalDateTime convert(String value) throws Exception {
            return LocalDateTime.parse(value, Dispatcher.DATE_FORMAT);
        }
    }
}
