package simulation.logging;

import ch.qos.logback.classic.spi.LoggingEvent;
import ch.qos.logback.core.AppenderBase;
import drones.drone.Drone;
import drones.geo.Point;
import drones.geo.TubeStation;
import drones.messages.Message.TrafficCondition.Condition;
import simulation.I18N;

import javax.imageio.ImageIO;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.nio.file.Path;
import java.util.List;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.regex.Pattern;

/**
 * Combines logs to form a map of the flow route and tube stations
 *
 * @author John Ahlroos
 */
public class RouteMapGenerator<E> extends AppenderBase<E> {

    private static final int IMAGE_PIXEL_SIZE = 800;
    private static final String DRONE_LOGGER = "Drone";
    private static final String DISPATCHER_LOGGER = "Dispatcher";
    private static final String REPORT_LOGGER = "Report";
    private static final String REPORT_FILE_NAME = "traffic-report.png";

    private static final Pattern ALL_TUBE_STATIONS_PATTERN = Pattern.compile(
            "Registered\sTube\sStation.*(name=(.*),\\slat=([\\d+\\.]+)),\\s(lon=([\\d+\\.-]+))");
    private static final Pattern VISITED_TUBE_STATIONS_PATTERN = Pattern.compile(
            "(.*) @ (\\d\\d:\\d\\d:\\d\\d): (.*) \\(drone: (\\d*),speed: (\\d*)km/h, distanceToStation: (\\d*)m\\)");
    private static final Pattern ROUTE_POINT_PATTERN = Pattern.compile(
            "Arrived.*(lat=([\\d+\\.]+)),\\s(lon=([\\d+\\.-]+))");

    private double min_lat = Double.MAX_VALUE;
    private double min_lon = Double.MAX_VALUE;
    private double max_lat = Double.MIN_VALUE;
    private double max_lon = Double.MIN_VALUE;

    private final Map<Long, List<Point>> points = new ConcurrentHashMap<>();
    private final List<TubeStation> stations = new ArrayList<>();
    private final Set<TubeStationWithCondition> foundStations = new HashSet<>();
    private final List<Long> terminated = new ArrayList<>();
    private boolean mapRendered = false;

    @Override
    protected void append(E event) {
        if (simulationTerminated()) {
            if (!mapRendered) {
                var file = renderMap();
                System.out.println();
                System.out.println(String.format(I18N.get("map.generation.message"), file.toAbsolutePath()));
                System.out.println();
            }
            return;
        }

        var loggingEvent = (LoggingEvent) event;
        var message = loggingEvent.getFormattedMessage();
        var loggerName = loggingEvent.getLoggerName();

        if (DISPATCHER_LOGGER.equals(loggerName)) {
            collectAllTubeStation(message);
            return;
        }

        if (REPORT_LOGGER.equals(loggerName)) {
            collectVisitedTubeStation(message);
            return;
        }

        if (loggerName.startsWith(DRONE_LOGGER)) {
            var droneId = Long.parseLong(loggerName.substring(loggerName.indexOf(".") + 1));
            registerDroneTerminated(droneId, message);
            collectRoutePoint(droneId, message);
            return;
        }
    }

    private void collectAllTubeStation(String message) {
        var m2 = ALL_TUBE_STATIONS_PATTERN.matcher(message);
        while (m2.find()) {
            var name = m2.group(2);
            var lat = Double.parseDouble(m2.group(3));
            var lon = Double.parseDouble(m2.group(5));
            var tb = new TubeStation(new Point(lat, lon, null), name);
            stations.add(tb);
        }
    }

    private void collectVisitedTubeStation(String message) {
        var m = VISITED_TUBE_STATIONS_PATTERN.matcher(message);
        while (m.find()) {
            var name = m.group(1);
            var time = m.group(2);
            var condition = m.group(3);
            var droneId = Long.parseLong(m.group(4));
            var speed = Integer.parseInt(m.group(5));
            stations.stream()
                    .filter(tb -> tb.name().equals(name))
                    .map(tb -> new TubeStationWithCondition(tb, time, condition, speed, droneId))
                    .forEach(foundStations::add);
        }
    }

    private void collectRoutePoint(long droneId, String message) {
        var m = ROUTE_POINT_PATTERN.matcher(message);
        while (m.find()) {
            var lat = Double.parseDouble(m.group(2));
            min_lat = Math.min(min_lat, lat);
            max_lat = Math.max(max_lat, lat);
            var lon = Double.parseDouble(m.group(4));
            min_lon = Math.min(min_lon, lon);
            max_lon = Math.max(max_lon, lon);
            points.putIfAbsent(droneId, new ArrayList<>());
            points.get(droneId).add(new Point(lat, lon, null));
        }
    }

    private void registerDroneTerminated(long droneId, String message) {
        if (message.contains("state=TERMINATED")) {
            terminated.add(droneId);
        }
    }

    private boolean simulationTerminated() {
        return !points.keySet().isEmpty() && terminated.containsAll(points.keySet());
    }

    private Path renderMap() {
        mapRendered = true;

        var colors = new ArrayDeque<>(List.of(Color.BLUE, Color.ORANGE));
        var map = new BufferedImage(IMAGE_PIXEL_SIZE, IMAGE_PIXEL_SIZE, BufferedImage.TYPE_INT_ARGB);
        var graphics = (Graphics2D) map.getGraphics();

        var font1 = new Font("Arial", Font.ITALIC, 18);
        var font2 = new Font("Arial", Font.BOLD, 16);
        var font3 = new Font("Arial", Font.PLAIN, 14);

        // Apply Zoom factor
        min_lon -= 0.006;
        max_lon += 0.006;
        min_lat -= 0.006;
        max_lat += 0.006;

        renderBackground(graphics);
        stations.forEach(tb -> renderStation(tb, graphics));
        foundStations.forEach(tb -> renderFoundStation(tb, graphics, font2, font3));
        points.forEach((droneId, points) -> renderRoute(droneId, points, graphics, font1, font2, colors));

        try {
            var file = Path.of(REPORT_FILE_NAME);
            ImageIO.write(map, "png", file.toFile());
            return file;
        } catch (IOException e) {
            System.err.println(I18N.get("map.file.generation.failed"));
        }
        return null;
    }

    private void renderBackground(Graphics graphics) {
        graphics.setColor(Color.LIGHT_GRAY);
        graphics.fillRect(0, 0, IMAGE_PIXEL_SIZE - 1, IMAGE_PIXEL_SIZE - 1);
    }

    private void renderStation(TubeStation tb, Graphics graphics) {
        if (foundStations.stream().map(TubeStationWithCondition::tubeStation).anyMatch(tb::equals)) {
            return;
        }
        var x = getXCoordinate(tb.point());
        var y = getYCoordinate(tb.point());
        graphics.setColor(new Color(170, 170, 170));
        graphics.fillRect(x - 5, y - 5, 10, 10);
        graphics.drawString(tb.name(), x + 10, y + 5);
    }

    private void renderFoundStation(TubeStationWithCondition tb, Graphics graphics, Font font2, Font font3) {
        var x = getXCoordinate(tb.tubeStation.point());
        var y = getYCoordinate(tb.tubeStation.point());

        graphics.setColor(switch (Condition.valueOf(tb.condition)) {
            case LIGHT -> Color.GREEN;
            case MODERATE -> Color.YELLOW;
            case HEAVY -> Color.RED;
        });

        graphics.setFont(font2);
        graphics.fillRect(x - 5, y - 5, 10, 10);
        graphics.drawString(String.format("%s@%s %s", tb.tubeStation.name(), tb.time, tb.condition), x + 10, y + 5);
        graphics.setColor(Color.DARK_GRAY);
        graphics.setFont(font3);
        graphics.drawString(String.format("(%s@%skm/h)", tb.droneId, tb.speed), x + 10, y + 20);
        graphics.setFont(font2);
    }

    private void renderRoute(long droneId, List<Point> data, Graphics graphics, Font font1, Font font2,
                             Deque<Color> colors) {

        // Convert range to pixels
        var range = data.get(0).moveTowards(data.get(1), Drone.TUBE_STATION_RANGE, 1);
        var rangeX = Math.abs(getXCoordinate(data.get(0)) - getXCoordinate(range));
        var rangeY = Math.abs(getYCoordinate(data.get(0)) - getYCoordinate(range));
        var rangeW = Math.max(rangeX, rangeY);

        var routeColor = colors.poll();
        var prevPoint = data.get(0);

        var prevX = getXCoordinate(prevPoint);
        var prevY = getYCoordinate(prevPoint);

        for (int i = 0; i < data.size(); i++) {
            var point = data.get(i);
            var x = getXCoordinate(point);
            var y = getYCoordinate(point);

            var rangeAlpha = (Math.abs(x - prevX + y - prevY) + 255) % 255;
            var rangeColor = new Color(0, 30, 254, rangeAlpha);

            graphics.setColor(rangeColor);
            graphics.fillOval(prevX - rangeW / 2, prevY - rangeW / 2, rangeW, rangeW);

            graphics.setColor(routeColor);
            graphics.drawLine(prevX, prevY, x, y);

            if (i == 0) {
                graphics.setFont(font1);
                graphics.fillOval(x - 5, y - 5, 10, 10);
                graphics.drawString(String.format(I18N.get("map.drone.start"), droneId), x - 20, y - 10);
                graphics.setFont(font2);
            } else if (i == data.size() - 1) {
                graphics.setFont(font1);
                graphics.fillOval(x - 5, y - 5, 10, 10);
                graphics.drawString(String.format(I18N.get("map.drone.end"), droneId), x - 20, y + 30);
                graphics.setFont(font2);
            }

            prevPoint = point;
            prevX = getXCoordinate(prevPoint);
            prevY = getYCoordinate(prevPoint);
        }
    }

    private int getXCoordinate(Point p) {
        var lonExtent = max_lon - min_lon;
        return (int) ((IMAGE_PIXEL_SIZE * (p.longitude() - min_lon)) / lonExtent);
    }

    private int getYCoordinate(Point p) {
        var latExtent = max_lat - min_lat;
        return (int) ((IMAGE_PIXEL_SIZE * (p.latitude() - min_lat)) / latExtent);
    }

    private record TubeStationWithCondition(
            TubeStation tubeStation, String time, String condition, int speed, long droneId) {

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;
            var that = (TubeStationWithCondition) o;
            return Objects.equals(tubeStation.name(), that.tubeStation.name());
        }

        @Override
        public int hashCode() {
            return Objects.hash(tubeStation.name());
        }
    }

}
